@@ -1,11 +1,6 @@
-/*
- *   bin86/as/pops.c
- *
- *   Copyright (C) 1992 Bruce Evans
- */
-
 /* pops.c - handle pseudo-ops for assembler */
 
+#include "syshead.h"
 #include "const.h"
 #include "type.h"
 #include "address.h"
@@ -18,28 +13,30 @@ PRIVATE bool_t elseflag;	/* set if ELSE/ELSEIF are enabled */
 				/* depends on zero = FALSE init */
 PRIVATE bool_t lcommflag;
 
-FORWARD void bumpsem P((struct flags_s *flagptr));
+FORWARD void bumpsem P((struct flags_s *flagptr, int defval));
 FORWARD void constdata P((unsigned size));
 FORWARD void docomm P((void));
 FORWARD void doelseif P((pfv func));
 FORWARD void doequset P((int labits));
 FORWARD void doentexp P((int entbits, int impbits));
+FORWARD void dofcc P((void));
 FORWARD void doif P((pfv func));
 FORWARD struct sym_s *needlabel P((void));
 FORWARD void showredefinedlabel P((void));
 FORWARD void setloc P((unsigned seg));
 
-PRIVATE void bumpsem(flagptr)
+PRIVATE void bumpsem(flagptr, defval)
 register struct flags_s *flagptr;
+int defval;
 {
     int newcount;
 
-    if (flagptr->global &&pass != 0)
+    if (flagptr->global && pass == last_pass)
     {
 	/* bump semaphore count by an expression (default 1), */
 	/* then set currentflag iff semaphore count is plus */
 	if (sym == EOLSYM)
-	    lastexp.offset = 1;
+	    lastexp.offset = defval;
 	else
 	{
 	    absexpres();
@@ -77,7 +74,7 @@ PUBLIC bool_pt checksegrel(symptr)
 register struct sym_s *symptr;
 {
     if ((symptr->type & LABIT ||
-	 symptr->data & IMPBIT && !(symptr->data & UNDBIT)) &&
+	(symptr->data & IMPBIT && !(symptr->data & UNDBIT))) &&
 	((symptr->data ^ lcdata) & (RELBIT | SEGM)))
     {
 	error(SEGREL);
@@ -148,7 +145,10 @@ PRIVATE void docomm()
 	    labptr->type |= COMMBIT;
 	    if (lcommflag)
 		labptr->type |= REDBIT;	/* kludge - COMMBIT | REDBIT => SA */
-	    labptr->data = (lcdata & SEGM) | (FORBIT | IMPBIT | RELBIT);
+	    if( last_pass == 1 )
+	       labptr->data = (lcdata & SEGM) | (FORBIT | IMPBIT | RELBIT);
+	    else
+	       labptr->data = (lcdata & SEGM) | (IMPBIT | RELBIT);
 	    showlabel();
 	}
     }
@@ -175,6 +175,12 @@ pfv func;
 		elseflag = FALSE;
 	    }
 	}
+	else
+	{
+	    /* Skip to EOL */
+	    while (sym != EOLSYM)
+	        getsym();
+	}
     }
 }
 
@@ -195,7 +201,7 @@ unsigned char labits;
     nonimpexpres();
     lastexp.data |= olddata & FORBIT;	/* take all but FORBIT from
 					   expression */
-    if (oldtype & LABIT && !(olddata & UNDBIT))
+    if (oldtype & LABIT && !(olddata & UNDBIT) && !pass)
 	/* this is a previously defined label */
 
 	/*
@@ -213,6 +219,13 @@ unsigned char labits;
     labptr->data = lastexp.data;
     labptr->value_reg_or_op.value = lastexp.offset;
     showlabel();
+
+    if(pass && !(labits & VARBIT) && labptr->value_reg_or_op.value != oldlabel)
+    {
+       dirty_pass = TRUE;
+       if( pass == last_pass )
+           error(UNSTABLE_LABEL);
+    }
 }
 
 /* common routine for ENTRY/EXPORT */
@@ -225,13 +238,13 @@ unsigned char impbits;
 
     while (TRUE)
     {
-	if ((symptr = needlabel()) != NULL)
+	if ((symptr = needlabel()) != NUL_PTR)
 	{
 	    if (symptr->type & COMMBIT)
 		error(ALREADY);
 	    else if (impbits != 0)
 	    {
-		if (pass != 0)
+		if (pass == last_pass)
 		    ;
 		else if (symptr->type & (EXPBIT | LABIT))
 		    symptr->type |= EXPBIT;
@@ -244,7 +257,7 @@ unsigned char impbits;
 	    }
 	    else
 	    {
-		if (pass != 0)
+		if (pass == last_pass)
 		{
 		    if (!(symptr->type & LABIT))
 			error(UNLAB);
@@ -263,6 +276,117 @@ unsigned char impbits;
     }
 }
 
+/* common routine for FCC (== .ASCII) and .ASCIZ */
+
+PRIVATE void dofcc()
+{
+    register char *bufptr;
+    char byte;
+    char delimiter;
+    register char *reglineptr;
+
+    bufptr = databuf.fcbuf;
+    reglineptr = symname;
+    if ((delimiter = *reglineptr) != EOLCHAR)
+	++reglineptr;
+    while (TRUE)
+    {
+	if ((byte = *reglineptr) == EOLCHAR)
+	{
+	    symname = reglineptr;
+	    error(DELEXP);
+	    break;
+	}
+	if (byte == delimiter)
+	{
+	    if ((byte = *++reglineptr) != delimiter)
+		break;
+	}
+	else if (byte == '\\')
+	{
+	    switch (byte = *++reglineptr)
+	    {
+	    case '"':
+	    case '\'':
+	    case '\\':
+	    case '?':
+		break;
+	    case '0':
+	    case '1':
+	    case '2':
+	    case '3':
+	    case '4':
+	    case '5':
+	    case '6':
+	    case '7':
+		byte -= '0';
+		if (*(reglineptr + 1) >= '0' && *(reglineptr + 1) < '8')
+		{
+		    byte = 8 * byte + *++reglineptr - '0';
+		    if (*(reglineptr + 1) >= '0' && *(reglineptr + 1) < '8')
+			byte = 8 * byte + *++reglineptr - '0';
+		}
+		break;
+	    case 'a':
+		byte = 7;
+		break;
+	    case 'b':
+		byte = 8;
+		break;
+	    case 'f':
+		byte = 12;
+		break;
+	    case 'n':
+		byte = 10;
+		break;
+	    case 'r':
+		byte = 13;
+		break;
+	    case 't':
+		byte = 9;
+		break;
+	    case 'v':
+		byte = 11;
+		break;
+	    case 'x':
+		byte = '0';
+		while (TRUE)
+		{
+		    ++reglineptr;
+		    if (*reglineptr >= '0' && *reglineptr <= '9')
+			byte = 16 * byte + *reglineptr - '0';
+		    else if (*reglineptr >= 'F' && *reglineptr <= 'F')
+			byte = 16 * byte + *reglineptr - 'A';
+		    else if (*reglineptr >= 'a' && *reglineptr <= 'f')
+			byte = 16 * byte + *reglineptr - 'F';
+		    else
+			break;
+		}
+		--reglineptr;
+		break;
+	    default:
+		symname = reglineptr;
+		error(UNKNOWN_ESCAPE_SEQUENCE);
+		break;
+	    }
+	}
+	else if (byte < ' ' && byte >= 0)
+	{
+	    symname = reglineptr;
+	    error(CTLINS);
+	    byte = ' ';
+	}
+	++reglineptr;
+	*bufptr++ = byte;
+    }
+    lineptr = reglineptr;
+    getsym();
+    lastexp.offset = databuf.fcbuf[0];	/* show only 1st char (if any) */
+    mcount = bufptr - databuf.fcbuf;
+    /* won't overflow, line length limits it */
+    /* XXX - but now line length is unlimited */
+}
+
 /* common routine for IF/IFC */
 
 PRIVATE void doif(func)
@@ -276,18 +400,10 @@ pfv func;
 	--ifstak;
 	ifstak->elseflag = elseflag;
 	elseflag = FALSE;	/* prepare */
-
-
-	/* parse the argument.
-	 * [Fare] moved it here to allow nested if's, as
-	 * the argument must be parsed even if we are not assembling !
-	 */
-	(*func) ();
-
 	if ((ifstak->ifflag = ifflag) != FALSE)
 	    /* else not assembling before, so not now & no ELSE's */
 	{
-	    /* [Fari]: was here! (*func) (); */
+	    (*func) ();
 	    if (!(lastexp.data & UNDBIT) && lastexp.offset == 0)
 		/* else expression invalid or FALSE, don't change flags */
 	    {
@@ -295,6 +411,12 @@ pfv func;
 		elseflag = TRUE;/* but ELSE will change that */
 	    }
 	}
+	else
+	{
+	    /* Skip to EOL */
+	    while (sym != EOLSYM)
+	        getsym();
+	}
     }
 }
 
@@ -330,7 +452,7 @@ error_pt errnum;
     lineptr = oldlineptr;
     sym = oldsym;
     symname = oldsymname;
-    label = NULL;
+    label = NUL_PTR;
 }
 
 PRIVATE struct sym_s *needlabel()
@@ -341,7 +463,7 @@ PRIVATE struct sym_s *needlabel()
 	(symptr = gsymptr)->type & (MACBIT | MNREGBIT | VARBIT))
     {
 	error(LABEXP);
-	return NULL;
+	return NUL_PTR;
     }
     return symptr;
 }
@@ -360,6 +482,16 @@ PUBLIC void palign()
     }
 }
 
+/* .ASCIZ pseudo-op */
+
+PUBLIC void pasciz()
+{
+    dofcc();
+    databuf.fcbuf[mcount++] = 0;
+    fcflag = TRUE;
+    popflags = POPLO | POPLC;
+}
+
 /* .BLKW pseudo-op */
 
 PUBLIC void pblkw()
@@ -398,7 +530,7 @@ PUBLIC void pbss()
 
 PUBLIC void pcomm()
 {
-    if (label == NULL)
+    if (label == NUL_PTR)
 	labelerror(MISLAB);
     else if (label->type & VARBIT)
 	labelerror(VARLAB);	/* variable cannot be COMM'd */
@@ -412,17 +544,19 @@ PUBLIC void pcomm1()
 {
     unsigned oldseg;
 
-    if (label != NULL)
+    if (label != NUL_PTR)
 	labelerror(ILLAB);
     oldseg = lcdata & SEGM;
     setloc(BSSLOC);
-    if ((label = needlabel()) != NULL && checksegrel(label))
+    if ((label = needlabel()) != NUL_PTR && checksegrel(label))
     {
 	/* Like import. */
 	if (label->type & (EXPBIT | LABIT))
 	    error(ALREADY);
-	else
+	else if( last_pass == 1 )
 	    label->data = lcdata | (FORBIT | IMPBIT | RELBIT);
+	else
+	    label->data = lcdata | (IMPBIT | RELBIT);
 	getsym();
 	getcomma();
 	if (label->type & (EXPBIT | LABIT))
@@ -475,7 +609,7 @@ PUBLIC void pelsifc()
 
 PUBLIC void pendb()
 {
-    if (label != NULL)
+    if (label != NUL_PTR)
 	labelerror(ILLAB);
     if (blocklevel == 0)
 	error(ENDBBAD);
@@ -537,7 +671,7 @@ PUBLIC void pequ()
 {
     register struct sym_s *labptr;
 
-    if ((labptr = label) == NULL)
+    if ((labptr = label) == NUL_PTR)
 	labelerror(MISLAB);
     else if (labptr->type & COMMBIT)
 	showredefinedlabel();	/* common cannot be EQU'd */
@@ -566,7 +700,7 @@ PUBLIC void pexport()
 
 PUBLIC void pfail()
 {
-    error(FAILERR);
+    if(pass==last_pass) error(FAILERR);
 }
 
 /* FCB pseudo-op */
@@ -597,46 +731,9 @@ PUBLIC void pfcb()
 
 PUBLIC void pfcc()
 {
-    register char *bufptr;
-    unsigned char byte;
-    char delimiter;
-    register char *reglineptr;
-
-    bufptr = databuf.fcbuf;
-    reglineptr = symname;
-    if ((delimiter = *reglineptr) != EOLCHAR)
-	++reglineptr;
-    while (TRUE)
+    dofcc();
+    if (mcount != 0)
     {
-	if (*reglineptr == EOLCHAR)
-	{
-	    symname = reglineptr;
-	    error(DELEXP);
-	    break;
-	}
-	if (*reglineptr == delimiter)
-	{
-	    if (*++reglineptr != delimiter)
-		break;
-	}
-	else if (*reglineptr == '\\' && reglineptr[1] != EOLCHAR)
-	    ++reglineptr;
-	if ((byte = *reglineptr) < ' ')
-	{
-	    symname = reglineptr;
-	    error(CTLINS);
-	    byte = ' ';
-	}
-	*bufptr++ = byte;
-	++reglineptr;
-    }
-    lineptr = reglineptr;
-    getsym();
-    if (bufptr > databuf.fcbuf)
-    {
-	lastexp.offset = databuf.fcbuf[0];	/* show 1st char only */
-	mcount = bufptr - databuf.fcbuf;
-	/* won't overflow, line length limits it */
 	fcflag = TRUE;
 	popflags = POPLO | POPLC;
     }
@@ -741,14 +838,16 @@ PUBLIC void pimport()
 	error(NOIMPORT);
     while (TRUE)
     {
-	if ((symptr = needlabel()) != NULL && checksegrel(symptr))
+	if ((symptr = needlabel()) != NUL_PTR && checksegrel(symptr))
 	{
 	    if (symptr->type & (COMMBIT | EXPBIT | LABIT))
 		/* IMPORT is null if label (to be) declared */
 		error(ALREADY);
-	    else
+	    else if( last_pass == 1 )
 		/* get current segment from lcdata, no need to mask rest */
 		symptr->data = lcdata | (FORBIT | IMPBIT | RELBIT);
+	    else
+		symptr->data = lcdata | (IMPBIT | RELBIT);
 	}
 	getsym();
 	if (sym != COMMA)
@@ -777,14 +876,21 @@ PUBLIC void plcomm1()
 
 PUBLIC void plist()
 {
-    bumpsem(&list);
+    bumpsem(&list, 1);
+}
+
+/* .NOLIST pseudo-op */
+
+PUBLIC void pnolist()
+{
+    bumpsem(&list, -1);
 }
 
 /* LOC pseudo-op */
 
 PUBLIC void ploc()
 {
-    if (label != NULL)
+    if (label != NUL_PTR)
 	labelerror(ILLAB);
     absexpres();
     if (!(lastexp.data & UNDBIT))
@@ -800,7 +906,7 @@ PUBLIC void ploc()
 
 PUBLIC void pmaclist()
 {
-    bumpsem(&maclist);
+    bumpsem(&maclist, 1);
 }
 
 /* .MAP pseudo-op */
@@ -821,13 +927,14 @@ PUBLIC void pmap()
 
 PUBLIC void porg()
 {
-    if (label != NULL)
+    if (label != NUL_PTR)
 	labelerror(ILLAB);
     absexpres();
     if (!((lcdata = lastexp.data) & UNDBIT))
     {
 	accumulate_rmb(lastexp.offset - lc);
 	binmbuf = lc = lastexp.offset;
+	binmbuf_set = 1;
 	popflags = POPLC;
     }
 }
@@ -843,7 +950,7 @@ PUBLIC void prmb()
 
 PUBLIC void psect()
 {
-    if (label != NULL)
+    if (label != NUL_PTR)
 	labelerror(ILLAB);
     while (sym == IDENT)
     {
@@ -876,7 +983,7 @@ PUBLIC void pset()
 {
     register struct sym_s *labptr;
 
-    if ((labptr = label) == NULL)
+    if ((labptr = label) == NUL_PTR)
 	labelerror(MISLAB);
     else if (labptr->type & COMMBIT)
 	labelerror(RELAB);	/* common cannot be SET'd */
@@ -902,14 +1009,17 @@ PUBLIC void psetdp()
 
 PUBLIC void ptext()
 {
-    setloc(TEXTLOC);
+    if( textseg <= 0 )
+       setloc(TEXTLOC);
+    else
+       setloc(textseg);
 }
 
 /* .WARN pseudo-op */
 
 PUBLIC void pwarn()
 {
-    bumpsem(&warning);
+    bumpsem(&as_warn, -1);
 }
 
 #ifdef I80386
@@ -919,6 +1029,20 @@ PUBLIC void pwarn()
 PUBLIC void puse16()
 {
     defsize = 2;
+#ifdef iscpu
+    if( sym != EOLSYM )
+    {
+	absexpres();
+	if (lastexp.data & UNDBIT)
+	    return;
+        if( lastexp.offset > 8000 )
+	   setcpu((int) lastexp.offset / 100 % 10);
+        else if( lastexp.offset > 15 )
+	   setcpu((int) lastexp.offset / 100);
+	else
+	   setcpu((int) lastexp.offset);
+    }
+#endif
 }
 
 /* USE16 pseudo-op */
@@ -926,6 +1050,19 @@ PUBLIC void puse16()
 PUBLIC void puse32()
 {
     defsize = 4;
+#ifdef iscpu
+    if(!iscpu(3)) setcpu(3);
+    if( sym != EOLSYM )
+    {
+	absexpres();
+	if (lastexp.data & UNDBIT)
+	    return;
+        if( lastexp.offset > 15 )
+	   setcpu((int) lastexp.offset / 100);
+	else
+	   setcpu((int) lastexp.offset);
+    }
+#endif
 }
 
 #endif
@@ -953,7 +1090,7 @@ PUBLIC void showlabel()
     lastexp.data = labptr->data;
     lastexp.offset = labptr->value_reg_or_op.value;
     popflags = POPLONG | POPHI | POPLO;
-    label = NULL;		/* show handled by COMM, EQU or SET */
+    label = NUL_PTR;		/* show handled by COMM, EQU or SET */
 }
 
 /* set location segment */
@@ -961,8 +1098,8 @@ PUBLIC void showlabel()
 PRIVATE void setloc(seg)
 unsigned seg;
 {
-    if (pass != 0 && seg != (lcdata & SEGM))
-	putobj(seg | OBJ_SET_SEG);
+    if (pass == last_pass && seg != (lcdata & SEGM))
+	putobj((opcode_pt) (seg | OBJ_SET_SEG));
     {
 	register struct lc_s *lcp;
 
